/* packet-thing.c
 * Routines for the thing Facility's Pseudo-Protocol dissection
 *
 * Copyright 2004, Luis E. Garcia Ontanon <lego@webflies.org>
 *
 * $Id: packet-thing.c,v 1.1 2004/08/24 23:28:41 lego Exp $
 *
 * Ethereal - Network traffic analyzer
 * By Gerald Combs <gerald@ethereal.com>
 * Copyright 1998 Gerald Combs
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


/**************************************************************************
 * This is the pseudo protocol dissector for the thing module.          ***
 * It is intended for this to be just the user interface to the module. ***
 **************************************************************************/

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <stdio.h>
#include <string.h>
#include <glib.h>
#include <epan/packet.h>
#include <epan/strutil.h>
#include <epan/thing.h>
#include <epan/thing_util.h>
/* #include "tap.h" */
#include "prefs.h"

static int proto_thing = -1;
static gint ett_thing = -1;
static int hf_thing = -1;
static int hf_thing_leg = -1;
static int hf_thing_leg_attr = -1;
static int hf_thing_leg_pdu = -1;
static int hf_thing_leg_start_time = -1;
static int hf_thing_leg_stop_time = -1;
static int hf_thing_leg_last_time = -1;

static int hf_thing_ses = -1;
static int hf_thing_pdu = -1;
static int hf_thing_pdu_rel_time = -1;


static char* pref_thing_config_filename = "config.thing";
static gboolean pref_thing_is_enabled = FALSE;

void init_thing_dissector(void) {
	
	if ( ! pref_thing_is_enabled ) {
		disable_thing();
		return;
	}
	
	initialize_thing( pref_thing_config_filename );
	
}

void avpl_tree(AVPL* avpl,proto_tree* tree,tvbuff_t *tvb, int hfi) {
	AVPN* c;
	guint8* s;
	
	for ( c = avpl->null.next; c->avp; c = c->next) {
		
		s = avp_to_str(c->avp);
		
		if (hfi) {
			proto_tree_add_string_format(tree,hfi,tvb,0,0,s,"%s",s);
		} else {
			proto_tree_add_text(tree,tvb,0,0,"%s",s);
		}
	}
}

void thing_leg_tree(proto_tree* pdu_tree, tvbuff_t *tvb, thing_leg* leg);

void thing_ses_tree(proto_tree* leg_tree, tvbuff_t *tvb, thing_ses* ses) {
	proto_item *ses_item;
	proto_tree *ses_tree;
	proto_item *ses_leg_item;
	proto_tree *ses_leg_tree;
	proto_item *ses_attr_item;
	proto_tree *ses_attr_tree;
	thing_leg* ses_legs;
	
	//g_message("tghing_ses_tree: 1");
	ses_item = proto_tree_add_string(leg_tree,hf_thing_ses,tvb,0,0,ses->id);
	ses_tree = proto_item_add_subtree(ses_item,ett_thing);
	
	ses_attr_item = proto_tree_add_text(ses_tree,tvb,0,0,"Ses Attributes");
	ses_attr_tree = proto_item_add_subtree(ses_attr_item, ett_thing);

	//g_message("tghing_ses_tree: 2");
	avpl_tree(ses->avpl,ses_attr_tree,tvb,hf_thing);
	
	//g_message("tghing_ses_tree: 3");

	ses_leg_item = proto_tree_add_text(ses_tree,tvb,0,0,"Ses Legs");
	ses_leg_tree = proto_item_add_subtree(ses_leg_item, ett_thing);
	
	//g_message("tghing_ses_tree: 4");

	for (ses_legs = ses->legs; ses_legs; ses_legs = ses_legs->next) {
		thing_leg_tree(ses_leg_tree, tvb, ses_legs);
	}
	//g_message("thing_ses_tree: 5");
	
}

void thing_leg_tree(proto_tree* pdu_tree, tvbuff_t *tvb, thing_leg* leg) {
	proto_item *leg_item;
	proto_tree *leg_time_tree;
	proto_item *leg_time_item;
	proto_tree *leg_tree;
	proto_item *leg_pdu_item;
	proto_tree *leg_pdu_tree;
	proto_item *leg_attr_item;
	proto_tree *leg_attr_tree;
	thing_pdu* leg_pdus;
	float  rel_time;
	float  leg_time;

	leg_item = proto_tree_add_string(pdu_tree,hf_thing_leg,tvb,0,0,leg->id);
	leg_tree = proto_item_add_subtree(leg_item, ett_thing);
	
	leg_attr_item = proto_tree_add_text(leg_tree,tvb,0,0,"Leg Attributes");
	leg_attr_tree = proto_item_add_subtree(leg_attr_item, ett_thing);
	
	avpl_tree(leg->avpl,leg_attr_tree,tvb,hf_thing_leg_attr);
	
	leg_time_item = proto_tree_add_text(leg_tree,tvb,0,0,"Leg Times");
	leg_time_tree = proto_item_add_subtree(leg_time_item, ett_thing);
	
	proto_tree_add_float(leg_time_tree, hf_thing_leg_start_time, tvb, 0, 0, leg->start_time);
				
	if (leg->released) { 
		proto_tree_add_float(leg_time_tree, hf_thing_leg_stop_time, tvb, 0, 0, leg->release_time);
		if (leg->release_time != leg->last_time) {
			proto_tree_add_float(leg_time_tree, hf_thing_leg_last_time, tvb, 0, 0, leg->last_time); 
		}
	} else {
		proto_tree_add_float(leg_time_tree, hf_thing_leg_last_time, tvb, 0, 0, leg->last_time); 
	}		

	leg_pdu_item = proto_tree_add_text(leg_tree,tvb,0,0,"Leg PDUs");
	leg_pdu_tree = proto_item_add_subtree(leg_pdu_item, ett_thing);
	
	rel_time = leg_time = leg->start_time;
	
	for (leg_pdus = leg->pdus; leg_pdus; leg_pdus = leg_pdus->next) {
		if (leg_pdus->is_start) {
			proto_tree_add_uint_format(leg_pdu_tree,hf_thing_leg_pdu,
									   tvb,0,0,leg_pdus->framenum,
									   "Start PDU: in frame %i",
									   leg_pdus->framenum);
		} else if (leg_pdus->is_stop) {
			proto_tree_add_uint_format(leg_pdu_tree,hf_thing_leg_pdu,
									   tvb,0,0,leg_pdus->framenum,
									   "Stop PDU: in frame %i (%f : %f)",
									   leg_pdus->framenum,
									   leg_pdus->rel_time,
									   leg_pdus->rel_time-rel_time);
			
		} else if (leg_pdus->after_release) {
			proto_tree_add_uint_format(leg_pdu_tree,hf_thing_leg_pdu,
									   tvb,0,0,leg_pdus->framenum,
									   "After stop PDU: in frame %i (%f : %f)",
									   leg_pdus->framenum,
									   leg_pdus->rel_time,
									   leg_pdus->rel_time-rel_time);
		} else {
			proto_tree_add_uint_format(leg_pdu_tree,hf_thing_leg_pdu,
									   tvb,0,0,leg_pdus->framenum,
									   "PDU: in frame %i (%f : %f)",
									   leg_pdus->framenum,
									   leg_pdus->rel_time,
									   leg_pdus->rel_time-rel_time);
		}
		
		rel_time = leg_pdus->rel_time;
		
	}
	
	/* if ( leg->ses ) thing_ses_tree(leg_tree,tvb,leg->ses); */
}


void thing_pdu_tree(thing_pdu *pdu, tvbuff_t *tvb, proto_tree* tree) {
	proto_item *pdu_item;
	proto_tree *pdu_tree;
	proto_item *pdu_attr_item;
	proto_tree *pdu_attr_tree;
	
	guint32 len;
	
	if ( ! pdu ) return;
	
	len = pdu->end - pdu->start;
	pdu_item = proto_tree_add_text(tree,tvb,pdu->start,len,"%s",pdu->id);
	pdu_tree = proto_item_add_subtree(pdu_item, ett_thing);
	proto_tree_add_float(pdu_tree, hf_thing_pdu_rel_time, tvb, 0, 0, pdu->rel_time);		

	if (pdu->leg) {
		if (pdu->leg->ses) {
			proto_tree_add_text(pdu_tree,tvb,0,0,"This Leg: %s",pdu->leg->id);
			thing_ses_tree(pdu_tree,tvb,pdu->leg->ses);
		} else {
			thing_leg_tree(pdu_tree,tvb,pdu->leg);
		}
	}
	
	if (pdu->avpl) {
		pdu_attr_item = proto_tree_add_text(pdu_tree,tvb,0,0,"PDU Attributes");
		pdu_attr_tree = proto_item_add_subtree(pdu_attr_item, ett_thing);
		avpl_tree(pdu->avpl,pdu_attr_tree,tvb,hf_thing_pdu);
	}
	
}



extern void dissect_thing(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree) {
	thing_pdu* pdus;
	proto_item *thing_item;
	proto_tree *thing_tree;

	if (tree) {
		
		pdus = do_thing(tvb,tree,pinfo);
	
		if (pdus) {
			/* TODO: colorize */
			
			thing_item = proto_tree_add_uint_format(tree,hf_thing_leg_pdu,
													  tvb,0,0,
													  pinfo->fd->num,"thing Info");
			
			thing_tree = proto_item_add_subtree(thing_item, ett_thing);
			
			for ( ; pdus; pdus = pdus->next_in_frame) {
				thing_pdu_tree(pdus,tvb,thing_tree);			
			}
		}
	}
}

void proto_reg_handoff_thing(void);

void
proto_register_thing(void)
{                 
	static hf_register_info hf[] = {
	{ &hf_thing, {"thing", "thing",
		FT_STRING, BASE_NONE, NULL, 0,
		"Interface to thing Facility",HFILL }},
	{ &hf_thing_leg, {"Leg", "thing.leg",
		FT_STRING, BASE_NONE, NULL, 0,
		"A thing Leg",HFILL }},
	{ &hf_thing_ses, {"Session", "thing.session",
		FT_STRING, BASE_NONE, NULL, 0,
		"A thing Session",HFILL }},
	{ &hf_thing_pdu, {"PDU Attribute", "thing.pdu.attr",
		FT_STRING, BASE_NONE, NULL, 0,
		"An Attribute of a PDU",HFILL }},
	{ &hf_thing_leg_attr, {"Leg Attribute", "thing.leg.attr",
		FT_STRING, BASE_NONE, NULL, 0,
		"An Attribute of a Leg",HFILL }},
	{ &hf_thing_leg_pdu, {"Leg's PDU", "thing.leg.pdu",
		FT_FRAMENUM, BASE_DEC, NULL, 0,
		"a PDU that belongss to this leg",HFILL }},
	{ &hf_thing_pdu_rel_time, {"seconds since start", "thing.leg.pdu.time",
		FT_FLOAT, BASE_DEC, NULL, 0,
		"Seconds passed since the start of the leg or capture if the pdu is unnassigned",HFILL }},
	{ &hf_thing_leg_start_time, {"start time", "thing.leg.start_time",
		FT_FLOAT, BASE_DEC, NULL, 0,
		"Seconds passed since the begining of caputre to the start of this leg",HFILL }},
	{ &hf_thing_leg_stop_time, {"leg time", "thing.leg.duration",
		FT_FLOAT, BASE_DEC, NULL, 0,
		"Duration in seconds from start to stop of this leg",HFILL }},
	{ &hf_thing_leg_last_time, {"last pdu", "thing.leg.excess_duration",
		FT_FLOAT, BASE_DEC, NULL, 0,
		"Time passed between the start of this leg and the last pdu assigned to it",HFILL }},
	};

	static gint *ett[] = {
		&ett_thing,
	};

	module_t *thing_module;

	proto_thing = proto_register_protocol("The thing Thing", "thing", "thing");

	proto_register_field_array(proto_thing, hf, array_length(hf));

	proto_register_subtree_array(ett, array_length(ett));
	
	register_dissector("thing",dissect_thing,proto_thing);
	
	register_init_routine(init_thing_dissector);

	thing_module = prefs_register_protocol(proto_thing, proto_reg_handoff_thing);
	
	prefs_register_bool_preference(thing_module, "enable",
								   "Enable thing",
								   "Whether the thing Facility should be enabled"
								   "If it does it will slow down ethereal and "
								   "(ab)use a lot of memory",
								   &pref_thing_is_enabled);
	
	prefs_register_string_preference(thing_module, "config_filename",
									 "Configuration Filename",
									 "The name of the file containing the thing module's configuration",
									 &pref_thing_config_filename);
	/* thing_tap = register_tap("ectaf"); */
}


void
proto_reg_handoff_thing(void)
{
	1;
}  



