/* thing.c
* The Thing
*
* Copyright 2004, Luis E. Garcia Ontanon <luis.ontanon@gmail.com>
*
* $Id: thing.c,v 0.4 2004/11/14 03:28:47 lego Exp $
*
* Ethereal - Network traffic analyzer
* By Gerald Combs <gerald@ethereal.com>
* Copyright 1998 Gerald Combs
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
* 
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

/* TODO:
 + "chunky" things
 + fix debug_print levels
 - timers
    - on legs
    - on sess
    - free
 - triggered events
    - how
 + transformations
    + maps
    + better config
 + the tap (to do the math)
    + think!
    + a scripting language? which?
 + fix configuration
	+ leaks
	+ reinit crashes
 + rewrite interface to ethereal 
    + resolve the pdu-frame relationship issue
    + rewrite make_pdus into proto.c
    + who should call the do_thing()?
*/

#define _THING_DEBUGGING


#include "config.h"
#include <stdio.h>
#include <string.h>
#include <glib.h> 
#include "thing.h"
#include "proto.h"
#include "thing_util.h"
#include <epan/epan_dissect.h>




typedef struct _finfo_node finfo_node;

struct _finfo_node {
	/* TODO: this sucks has to be rewritten */ 
	field_info* finfo;
	finfo_node* next;
	finfo_node* prev;	
};

typedef struct _working_packet {
	/* TODO: this sucks has to be rewritten */ 
	guint32 framenum;
	guint32 t;
	guint32 ut;
	finfo_node transport;
	finfo_node pdu;
	finfo_node attr;	
} working_packet;


/*******************************************************
** config data (once initialized these won't change) **
*******************************************************/

/*
 *  these are user set preferences 
 */

/* set by dissector */
static gboolean thing_enabled_flag = TRUE; /* disabled by default */

/* set by Settings config AVPL */
static guint32 max_things = 0;

/* remove this */
static guint32 workspace_scale = 10;

static gboolean keep_pdu_attributes = TRUE;

/* a count of existing things */
static guint current_things = 0;

/* a user setting */
static float ses_expiration = 2.0;



/* key: the name of the action
   value: a pointer to an config_action */
static GHashTable* actions = NULL;

/*
 *  these are used to import interesting_hfids 
 *  their key is a pointer to an hfid
 *
 *  TODO: the whole inteerface to ethereal sucks I must rewrite it
 */

/*   key: an hfid
   value: the name of the transport as a string  */
static GHashTable* hfids_transport = NULL;

/*   key: an hfid
   value: the name of the pdu's proto as a string */
static GHashTable* hfids_pdu = NULL;

/*   key: an hfid
   value: a string pointing to the name the avp will take in the pdu */
static GHashTable* hfids_attr = NULL;

/* array of pointers to integer containing the hfid of all the fields we need to trace
   used for priming the tree */
static GPtrArray* thing_hfids;


/*
 * These are populated from the user config and
 * used by the thing to do its job
 */


/*   key: a leg type name
   value: avpl (op) of conditions on which to start the thing of a certain leg */
static GHashTable* starts = NULL;

/*   key: a leg type name
  value: avpl ops of conditions on which to stop the thing of a certain leg */
static GHashTable* stops = NULL;

/*   key: a leg or pdu type name
   value: avpl ops of keys to legs (from pdu) */
static GHashTable* legkeys = NULL;

/*   key: a leg or pdu type name
   value: loal of ops of keys to sess (from leg) */
/* the last one reversed */
static GHashTable* seskeys = NULL;

/* key: a ses type 
  value: loal of ops of keys to sess (from leg) */
static GHashTable* syekses = NULL;


/*   key: a leg or pdu type name
value:  avpl ops of attributes to add to legs (from pdu) or sess (from leg) */
static GHashTable* extras = NULL;

/*    key: a leg or pdu type name
	value: GPtrArray of avpl_transfs */
static GHashTable* transforms = NULL;

/* the collection of single copy strings */
static GHashTable* strings = NULL;

/******************************************************************
 ** runtime data  (these will contain the info needed for thing) **
 ******************************************************************/

/* key: a thing id
   value: a pointer to the thing */
/* Is this realy needed or is it just a waste of memory? */
static GHashTable* things = NULL;

/*   key: a pointer to the integer representing the frame number
   value: a list of pdus contained within (NULL if none)*/
static GHashTable* visited_packets = NULL;

/*   key: a string representing an avpl which is key to a leg
   value: a ses */
static GHashTable* sessions = NULL;

/*   key: a leg or pdu type name
   value: a pointer to an integer count of pdus of that type (for use by pdu_id) */
static GHashTable* sessids = NULL;


/* head/tail of the "expirations" queue/list */
static thing circle_of_life;
static gboolean circle_of_life_initialized = FALSE;

/* set by make_pdus used to arbitrate session expiration */
static float now = 0.0;

/* is this realy used? */
static guint last_visited_frame = 0;


/* DEBUGGING settings */
#ifdef _THING_DEBUGGING

static int dbg_lvl = 0;
static int* dbg = &dbg_lvl;

static int dbg_leak_lvl = 0;
static int* dbg_leak = &dbg_leak_lvl;

static int dbg_cfg_lvl = 0;
static int* dbg_cfg = &dbg_cfg_lvl;

static int dbg_pdu_lvl = 0;
static int* dbg_pdu = &dbg_pdu_lvl;

static int dbg_leg_lvl = 0;
static int* dbg_leg = &dbg_leg_lvl;

static int dbg_ses_lvl = 0;
static int* dbg_ses = &dbg_ses_lvl;

FILE* dbg_facility = NULL;

#endif /* _THING_DEBUGGING */

/********* functions **********/

/*
 *  I delete a thing, if the thing has a parent I delegate the parent to delete
 *  the whole family
 */
gboolean delete_thing(thing* it) {
	thing* curr;
	guint i;
	void *k;
	void *v;
	
	current_things--;
	g_hash_table_remove(things,it->id);
	g_free(it->id);
#ifdef _THING_DEBUGGING
	dbg_print (dbg_leak,2,dbg_facility,"delete_thing: free: %X\n",it->id);
#endif
	
	if (it->avpl) {
		delete_avpl(it->avpl,TRUE);
#ifdef _THING_DEBUGGING
		dbg_print (dbg_leak,2,dbg_facility,"delete_thing: delete_avpl: %X\n",it->avpl);
#endif
	}
	
	if ( it->type == THING_SES_TYPE ) {
		/* a ses */
		for (curr = it->legs; curr; curr = curr->next) { 
			curr->ses = NULL;
			delete_thing(curr);
		}
		
		for (i = 0; i < it->ses_keys->len; i++) {
			g_hash_table_lookup_extended(sessions,g_ptr_array_index(it->ses_keys,i),&k,&v);
			g_hash_table_remove(sessions,k);
			g_free(k);
#ifdef _THING_DEBUGGING
			dbg_print (dbg_leak,2,dbg_facility,"delete_thing: free: %X\n",k);
#endif
		}
		
	} else if (it->type == THING_PDU_TYPE) {
		/* a leg */
		if (it->ses) {
			return delete_thing(it->ses);
		}
		
		for (curr = it->pdus; curr; curr = curr->next) { 
			curr->leg = NULL;
			delete_thing(curr);
		}
		
		g_hash_table_lookup_extended(sessions,it->leg_key,&k,&v);
		g_hash_table_remove(sessions,k);
		g_free(it->leg_key);
#ifdef _THING_DEBUGGING
		dbg_print (dbg_leak,2,dbg_facility,"delete_thing: free: %X\n",it->leg_key);
#endif
	} else {
		/* a pdu */
		if (it->leg) {
			return delete_thing(it->leg);
		}
		
		if (it->first) { 
			/* FIXME: I should try to delete every pdu in the frame! */
			g_hash_table_lookup_extended(visited_packets,&it->framenum,&k,&v);
			g_hash_table_remove(visited_packets,k);
			g_free(k);
#ifdef _THING_DEBUGGING
			dbg_print (dbg_leak,2,dbg_facility,"delete_thing: free: %X\n",k);
#endif
			
		}
	}

	
	if ( it->next_in_ring ) {
		it->next_in_ring->prev_in_ring = it->prev_in_ring;
		it->prev_in_ring->next_in_ring = it->next_in_ring;
	}
	
	g_free(it);
#ifdef _THING_DEBUGGING
	dbg_print (dbg_leak,2,dbg_facility,"delete_thing: free: %X\n",it);
#endif
	return TRUE;
}

/* generates the next thing id for a certain kind of thing */
guint8* thing_id(guint8* kind) {
	int *ip = NULL;
	int i;
	guint8* the_id;
	
	ip = g_hash_table_lookup(sessids,kind);
	
	if (ip) {
		(*ip)++;
		i = *ip;
	} else {
		i = 1;
		ip = g_malloc(sizeof(int));
#ifdef _THING_DEBUGGING
		dbg_print (dbg_leak,2,dbg_facility,"thing_id: new: %X\n",ip);
#endif		
		*ip = i;
		g_hash_table_insert(sessids,kind,ip);
	}
	
	the_id = g_strdup_printf("%s:%i",kind,i);
	
#ifdef _THING_DEBUGGING
	dbg_print (dbg_leak,2,dbg_facility,"thing_id: new: %X\n",the_id);
#endif		
	
	return the_id;
}

/* creates a thing */
thing* new_thing(guint8* kind) {
	thing* it = g_malloc(sizeof(thing));
	
	it->id  = NULL ;
	it->avpl  = NULL ;
	it->kind  = NULL ;
	it->start  = 0 ;
	it->end  = 0 ;
	it->framenum  = 0 ;
	it->next  = NULL ;
	it->released  = FALSE ;
	it->expiration = 0.0;

#ifdef _THING_DEBUGGING
	dbg_print (dbg_leak,2,dbg_facility,"new_thing: new: %X\n",it);
#endif

#if 0	
	if (max_things && current_things >= max_things) {
		delete_thing(circle_of_life.prev_in_ring);
	}

	it->prev_in_ring = &circle_of_life;
	it->next_in_ring = circle_of_life.next_in_ring;
	circle_of_life.next_in_ring = it;
	it->next_in_ring->prev_in_ring = it;
#endif
	
	current_things++;
	it->id = thing_id(kind); 
	it->kind = scs_subscribe(strings,kind);	
	
	return it;
}

/* a new leg */
thing_leg* new_leg(thing_pdu* pdu, guint8* key) {
	thing_leg* leg = new_thing(pdu->kind);

#ifdef _THING_DEBUGGING
	dbg_print (dbg_leg,1,dbg_facility,"new_leg: %s: ``%s''",leg->id,key);
#endif
	
	leg->avpl = new_avpl("attributes");
	
	leg->type = THING_LEG_TYPE;
	leg->ses = NULL;
	leg->pdus = pdu;
	leg->last_pdu = pdu;
	leg->leg_key = key;
	leg->next = NULL;
	leg->start_time = pdu->rel_time;
	leg->release_time = 0.0;
	leg->last_time = 0.0;
	
	pdu->leg = leg;
	pdu->next = NULL;
	pdu->is_start = TRUE;
	pdu->rel_time = 0.0;
#if 0	
	if (pdu->next_in_ring) {
		//g_message("get the pdu out of the circle of life");
		
		pdu->next_in_ring->prev_in_ring = pdu->prev_in_ring;
		pdu->prev_in_ring->next_in_ring = pdu->next_in_ring;
		
		pdu->prev_in_ring = NULL;
		pdu->next_in_ring = NULL;
	}
#endif	
	return leg;
}


void adopt_leg(thing_ses* ses, thing_leg* leg) {
	
#ifdef _THING_DEBUGGING
	dbg_print (dbg_ses,5,dbg_facility,"adopt_leg: ses=%X leg=%X",ses,leg);
#endif
	
	leg->ses = ses;
	leg->next = NULL;
	
	ses->num_of_legs++;
	
	if (ses->last_leg) {		
		ses->last_leg->next = leg;
	}

	ses->last_leg = leg;
	
	if (! ses->legs ) {
		ses->legs = leg;
	}
#if 0	
	if (leg->next_in_ring) {
		/* get the leg out of the circle of life */
		leg->next_in_ring->prev_in_ring = leg->prev_in_ring;
		leg->prev_in_ring->next_in_ring = leg->next_in_ring;
		
		leg->prev_in_ring = NULL;
		leg->next_in_ring = NULL;
	}
#endif

}

/* a new ses */
thing_ses* new_ses(thing_leg* leg, guint8* type) {
	thing_ses* ses = new_thing(type);
	
#ifdef _THING_DEBUGGING
	dbg_print (dbg_ses,1,dbg_facility,"new_ses: %s for %s",ses->id,leg->id);
#endif
	
	ses->type = THING_SES_TYPE;
	ses->avpl = new_avpl("");
	ses->legs = NULL;
	ses->last_n = 0;
	ses->ses_keys = g_ptr_array_new();
	ses->last_leg = NULL;
	
#ifdef _THING_DEBUGGING
	dbg_print (dbg_leak,2,dbg_facility,"new_ses: new_ptr_array: %X\n",ses->ses_keys);
#endif
	
	ses->start_time = leg->rel_time;
	
	adopt_leg(ses,leg);
	
	return ses;
}

/* deletes every object */
void thing_flush()  {
	while(circle_of_life.next != &circle_of_life) {
		delete_thing(circle_of_life.next);
	}
}

extern gboolean thing_enabled() {
	return thing_enabled_flag;
}


extern void enable_thing() {
	thing_enabled_flag = TRUE;
}

extern void disable_thing() {
	thing_enabled_flag = FALSE;
#ifdef _THING_DEBUGGING
	dbg_print (dbg,2,dbg_facility,"disable_thing\n");
#endif
}

/* called while iterating the interestig_hfids hash */
void thing_import_finfos(gpointer k, gpointer v, gpointer p ) {
	/* TODO: rewrite the whole thing, It sucks! */

	guint8* transport;
	guint8* pdu;
	guint8* attr;	
	int id;
	field_info* finfo;
	finfo_node* finode;
	GPtrArray	*ptrs = (GPtrArray*) v;
	working_packet* wp = (working_packet*) p;
	guint i = 0;

#ifdef _THING_DEBUGGING
	dbg_print (dbg_pdu,2,dbg_facility,"thing_import_finfos: %X\n", v);
#endif

	k = k; /* no warning */
	
	if ( ! ptrs ) return;
	for ( ; i < ptrs->len ; i++) {
		finfo = g_ptr_array_index(ptrs,i);
		
		id = finfo->hfinfo->id;
		
#ifdef _THING_DEBUGGING
		dbg_print (dbg_pdu,2,dbg_facility,"thing_import_finfos: about to import: %s(%i)\n",finfo->hfinfo->abbrev,finfo->hfinfo->id);
#endif

		if (( transport = g_hash_table_lookup(hfids_transport,&id) )) {
#ifdef _THING_DEBUGGING
			dbg_print (dbg_pdu,3,dbg_facility,"thing_import_finfos: importing %i as transport\n",id);
#endif
			finode = g_malloc(sizeof(finfo_node));
			finode->finfo = finfo;
			finode->next = &wp->transport;
			finode->prev = wp->transport.prev;
			wp->transport.prev->next = finode;
			wp->transport.prev = finode;
		}
		
		if (( pdu = g_hash_table_lookup(hfids_pdu,&id) )) {
#ifdef _THING_DEBUGGING
			dbg_print (dbg_pdu,3,dbg_facility,"thing_import_finfos: importing %i as pdu\n",id);
#endif
			finode = g_malloc(sizeof(finfo_node));
			finode->finfo = finfo;
			finode->next = &wp->pdu;
			finode->prev = wp->pdu.prev;
			wp->pdu.prev->next = finode;
			wp->pdu.prev = finode;
		}
		
		if (( attr = g_hash_table_lookup(hfids_attr,&id) )) {
#ifdef _THING_DEBUGGING
			dbg_print (dbg_pdu,3,dbg_facility,"thing_import_finfos: importing %i as attribute\n",id);
#endif
			finode = g_malloc(sizeof(finfo_node));
			finode->finfo = finfo;
			finode->next = &wp->attr;
			finode->prev = wp->attr.prev;
			wp->attr.prev->next = finode;
			wp->attr.prev = finode;			
		}
	}
	
}

extern void extract_interesting_finfos(proto_tree *tree, working_packet* wp) {
	/* TODO: rewrite the whole thing, It sucks! */
#ifdef _THING_DEBUGGING
	dbg_print (dbg_pdu,6,dbg_facility,"extract_interesting_finfos: %X %X %X %X\n",
			   tree->tree_data,tree->tree_data->interesting_hfids,wp);
#endif
	
	g_hash_table_foreach(tree->tree_data->interesting_hfids,thing_import_finfos,wp);
}


#ifdef _THING_DEBUGGING
void print_flist(guint8* what, finfo_node* fn) {
	guint8* name;
	guint8* value;
	guint32 from;
	guint32 to;
	
	
	for (fn = fn->next; fn->finfo; fn = fn->next) {
		name = fn->finfo->hfinfo->abbrev;
		from = fn->finfo->start;
		to = from + fn->finfo->length;

		if (fn->finfo->value.ftype->get_value_integer) {
			value = g_strdup_printf("%i",fvalue_get_integer(&fn->finfo->value));
		} else if (fn->finfo->value.ftype->val_to_string_repr) {
			value = fvalue_to_string_repr(&fn->finfo->value,FTREPR_DISPLAY,NULL);
		} else if (fn->finfo->value.ftype->get_value_floating) {
			value = g_strdup_printf("%f",fvalue_get_floating(&fn->finfo->value));
		} else {
			value = g_strdup("");
		}
		dbg_print (dbg_pdu,3,dbg_facility,"%s: %s(%i-%i)=%s",what,name,from,to,value);
		g_free(value);
	}	
}

void print_wp(working_packet* wp) {
	
	if (*dbg_pdu < 1) return;
	
	print_flist("transport",&wp->transport);
	print_flist("pdu",&wp->pdu);
	print_flist("attr",&wp->attr);
}
#endif


/* makes all the pdus in the frame */ 
thing_pdu* make_pdus(working_packet* wp) {
	/* TODO: rewrite the whole thing, It sucks! */
	/* TODO: handle more transports for a pdu (ex: ip+tcp) */
	/* FIXME: q931.redirecting vs e164.calling */

	finfo_node* fn_pdu;
	finfo_node* fn_attr;
	finfo_node* next_attr;
	
	finfo_node* fn_transport;
	finfo_node* curr_transport;
	gint curr_transport_start;
	gint curr_transport_end;
	
	AVP* avp;
	
	
	gint attr_start;
	gint attr_end;
	gint transport_start;
	gint transport_end;
	thing_pdu *pdu;
	thing_pdu *last_pdu = NULL;
	thing_pdu *first_pdu = NULL;
	guint8* as;
	
#ifdef _THING_DEBUGGING
	dbg_print (dbg_pdu,5,dbg_facility,"make_pdus: entering\n");
#endif
		
	for (fn_pdu = wp->pdu.next; fn_pdu->finfo; fn_pdu = fn_pdu->next) {
		current_things++;

		pdu = new_thing("pdu");
		
#ifdef _THING_DEBUGGING
		dbg_print (dbg_pdu,5,dbg_facility,"make_pdus: pdu %X\n",pdu);
#endif
		
		/* global */
		now = (((float)wp->t) + (((float)wp->ut) / 1000000) );
		
		pdu->type = THING_PDU_TYPE;
		pdu->kind = scs_subscribe(strings,fn_pdu->finfo->hfinfo->abbrev);
		pdu->start = fn_pdu->finfo->start;
		pdu->end = pdu->start + fn_pdu->finfo->length;
		pdu->avpl = new_avpl(pdu->id);
		pdu->leg = NULL;
		pdu->next_in_frame = NULL;
		pdu->next = NULL;
		pdu->first = FALSE;
		pdu->framenum = wp->framenum;
		pdu->is_start = FALSE;
		pdu->is_stop = FALSE;
		pdu->after_release = FALSE;
		pdu->rel_time = now;

		if (last_pdu) {
			last_pdu->next_in_frame = pdu;
		}
		
		last_pdu = pdu;

		if (! first_pdu ) {
			first_pdu = pdu;
		} else {
			pdu->first = TRUE;
		}

		curr_transport_start = 0;
		curr_transport_end = 9999999; /* that's a huge packet! */
		
		for ( fn_transport = wp->transport.next;
			  fn_transport->finfo;
			  fn_transport = fn_transport->next) {
			transport_start = fn_transport->finfo->start;
			transport_end = transport_start + fn_transport->finfo->length;
			
			if (  fn_transport->finfo->start > curr_transport_start 
 				  && fn_transport->finfo->start < pdu->start ) {
				curr_transport = fn_transport;
				curr_transport_start = fn_transport->finfo->start;
				curr_transport_end = curr_transport_start + fn_transport->finfo->length;
			}
		}
		
		for (fn_attr = wp->attr.next; fn_attr->finfo; fn_attr = next_attr) {
			next_attr = fn_attr->next;
			
			attr_start = fn_attr->finfo->start;
			attr_end = attr_start + fn_attr->finfo->length;
			
			if (pdu->start <= attr_start && pdu->end >= attr_end ) {
				if ((as = g_hash_table_lookup(hfids_attr,&fn_attr->finfo->hfinfo->id))) {
					avp = new_avp_from_finfo(as, fn_attr->finfo);
					
					if ( ! insert_avp(pdu->avpl,avp) ) {
						delete_avp(avp);
					}
					
					fn_attr->next->prev = fn_attr->prev;
					fn_attr->prev->next = fn_attr->next;
					g_free(fn_attr);
				}
			}
			
			if (curr_transport_start <= attr_start &&  curr_transport_end >= attr_end ) {
				if ((as = g_hash_table_lookup(hfids_attr,&fn_attr->finfo->hfinfo->id))) {
					
					insert_avp(pdu->avpl,
							   new_avp_from_finfo(as, fn_attr->finfo));
				}
			}
		}
	}
	
	return first_pdu;
}


void cleanup_working_packet(working_packet* wp) {
	finfo_node *fn;
	finfo_node *next_fn;
		
	for (fn = wp->pdu.next; fn->finfo; fn = next_fn) {
		next_fn = fn->next;
		g_free(fn);
#ifdef _THING_DEBUGGING
		dbg_print (dbg_leak,2,dbg_facility,"cleanup_working_packet: free: %X\n",fn);
#endif
		
	}

	for (fn = wp->attr.next; fn->finfo; fn = next_fn) {
		next_fn = fn->next;
		g_free(fn);
#ifdef _THING_DEBUGGING
		dbg_print (dbg_leak,2,dbg_facility,"cleanup_working_packet: free: %X\n",fn);
#endif		
	}

	for (fn = wp->transport.next; fn->finfo; fn = next_fn) {
		next_fn = fn->next;
		g_free(fn);
#ifdef _THING_DEBUGGING
		dbg_print (dbg_leak,2,dbg_facility,"cleanup_working_packet: free: %X\n",fn);
#endif
	}
	
}

/* applies the transform list for which type to what avpl */
void apply_transforms(AVPL* what, guint8* which) {
	GPtrArray* transformations = NULL;
	AVPL_Transf* transform = NULL;
	guint i;
	
#ifdef _THING_DEBUGGING
	dbg_print (dbg_pdu,3,dbg_facility,"apply_transforms: entering: what='%s' which='%s'\n",what->name,which);
#endif
	
	if (( transformations = g_hash_table_lookup(transforms,which) )) {		
		//g_message("there is transform list");
		for (i = 0; i < transformations->len; i++) {	
			transform = g_ptr_array_index(transformations,i);
			avpl_transform(what, transform);
		}
	}
}

/* applies the extras for which type to what avpl */
void apply_extras(AVPL* from, AVPL* to,  guint8* which) {
	AVPL* candidate_extras = NULL;
	AVPL* our_extras = NULL;

	if (( candidate_extras = g_hash_table_lookup(extras,which) )) {
		
#ifdef _THING_DEBUGGING
		dbg_print (dbg_pdu,3,dbg_facility,"apply_extras: entering: from='%s' to='%s' which='%s'\n",from->name,to->name,which);
#endif
		
		our_extras = new_avpl_loose_match("",from, candidate_extras, FALSE) ;
		
#ifdef _THING_DEBUGGING
		dbg_print (dbg_pdu,6,dbg_facility,"apply_extras: check: %X\n",our_extras);
#endif
		
		if (our_extras) {
			merge_avpl(to,our_extras,TRUE);
			delete_avpl(our_extras,FALSE);
		}
	}
}

void ses_remove_keys (thing_ses* ses) {
	guint8* k;

	while (ses->ses_keys->len) {
		k = (guint8*) g_ptr_array_remove_index_fast(ses->ses_keys,0);
		g_hash_table_remove(sessions,k);
		g_free(k);
		
#ifdef _THING_DEBUGGING
		dbg_print (dbg_leak,2,dbg_facility,"ses_remove_keys: free: %X\n",k);
#endif		
		
	}
}

void reanalyze_leg(thing_leg* leg) {
	LoAL* ses_keys = NULL;
	AVPL* curr_seskey = NULL;
	void* cookie = NULL;
	AVPL* seskey_match = NULL;
	thing_ses* ses = leg->ses;
	LoAL* matching_seskeys = NULL;
	guint8* key;
	
	if ( ! ses ) return;
	
	
#ifdef _THING_DEBUGGING
	dbg_print (dbg_ses,1,dbg_facility,"reanalize_leg: leg=%s ses=%s\n",leg->id,ses->id);
#endif
	
	apply_extras(leg->avpl,ses->avpl,leg->kind);
	
	if (ses->last_n != ses->avpl->len) {
		
#ifdef _THING_DEBUGGING
		dbg_print (dbg_ses,2,dbg_facility,"analize_leg: ses has new attributes let's look for new keys\n");
#endif
		
		ses_keys = g_hash_table_lookup(seskeys,ses->kind);
		
		while (( curr_seskey = get_next_avpl(ses_keys,&cookie) )) {
			if (( seskey_match = new_avpl_exact_match("",ses->avpl,curr_seskey,FALSE) )) {
				key = avpl_to_str(seskey_match);
				
				if ( g_hash_table_lookup(sessions,key) ) {
					g_free(key);
				} else {
#ifdef _THING_DEBUGGING
					dbg_print (dbg_ses,1,dbg_facility,"analize_leg: new key for ses=%s : %s\n",ses->id,key);
#endif
					g_hash_table_insert(sessions,key,ses);
					g_ptr_array_add(ses->ses_keys,key);
				}
				
				delete_avpl(seskey_match,FALSE);
			}
		}
		
		ses->last_n = ses->avpl->len;
	}
	
	if (ses->num_of_released_legs == ses->num_of_legs) {
		ses->released =  TRUE;
		ses->expiration = ses_expiration + now;
	} else {
		ses->released =  FALSE;
	}
	
}

void analize_leg(thing_leg* leg) {
	LoAL* ses_keys = NULL;
	AVPL* curr_seskey = NULL;
	void* cookie = NULL;
	AVPL* seskey_match = NULL;
	thing_ses* ses = NULL;
	guint8* key;
#ifdef _THING_DEBUGGING
	guint8* avpl_str;
#endif		
	
	if ( ! ( ses = leg->ses ) ) {
		/* no ses, let's either find one or create it if due */
#ifdef _THING_DEBUGGING
		dbg_print (dbg_ses,1,dbg_facility,"analize_leg: no ses\n");
#endif		
		
		ses_keys = g_hash_table_lookup(syekses,leg->kind);
		
		
		if ( ! ses_keys ) {
#ifdef _THING_DEBUGGING
			dbg_print (dbg_ses,1,dbg_facility,"analize_leg: no ses_keys for this leg\n");
#endif		
			return;
		}
		
		/* We'll look for any matching seskeys */
		
#ifdef _THING_DEBUGGING
		dbg_print (dbg_ses,1,dbg_facility,"analize_leg: got ses_keys\n");
#endif		
		
		while (( curr_seskey = get_next_avpl(ses_keys,&cookie) )) {
			
#ifdef _THING_DEBUGGING
			avpl_str = avpl_to_str(curr_seskey);
			dbg_print (dbg_ses,2,dbg_facility,"analize_leg: about to match for %s\n",avpl_str);
			g_free(avpl_str);
#endif		

			if (( seskey_match = new_avpl_exact_match(curr_seskey->name,leg->avpl,curr_seskey,TRUE) )) {
				
				key = avpl_to_str(seskey_match);

#ifdef _THING_DEBUGGING
				dbg_print (dbg_leak,1,dbg_facility,"analize_leg: new: %X\n",key);
				dbg_print (dbg_ses,1,dbg_facility,"analize_leg: got seskey_match: %s\n",key);
#endif
				
				if (( ses = g_hash_table_lookup(sessions,key) )) {
#ifdef _THING_DEBUGGING
					dbg_print (dbg_ses,1,dbg_facility,"analize_leg: got already a matching ses\n");
#endif
					
					if (ses->num_of_legs == ses->num_of_released_legs && ses->expiration < now) {
#ifdef _THING_DEBUGGING
						dbg_print (dbg_ses,1,dbg_facility,"analize_leg: this is a new ses, not the old one, let's create it\n");
#endif
						
						ses_remove_keys(ses);

						ses = new_ses(leg,seskey_match->name);
						ses->num_of_legs = 1;
					} else {
#ifdef _THING_DEBUGGING
						dbg_print (dbg_ses,1,dbg_facility,"analize_leg: this is our ses\n");
#endif
						
						g_free(key);
						
#ifdef _THING_DEBUGGING
						dbg_print (dbg_leak,2,dbg_facility,"analize_leg: free: %X\n",key);
#endif
						if (! leg->ses ) adopt_leg(ses,leg);
					}
				} else {
					
					if (! ses ) {
#ifdef _THING_DEBUGGING
						dbg_print (dbg_ses,1,dbg_facility,"analize_leg: no such ses in hash, let's create a new one\n");
#endif
					
						ses = new_ses(leg,seskey_match->name);
						ses->num_of_legs = 1;
					} else {
#ifdef _THING_DEBUGGING
						dbg_print (dbg_ses,1,dbg_facility,"analize_leg: no such ses in hash but we've got one let's merge\n");
#endif
						merge_avpl(ses->avpl,seskey_match,TRUE);
						
					}
				}
				
				delete_avpl(seskey_match,TRUE);
				
			}
#ifdef _THING_DEBUGGING
			dbg_print (dbg_ses,1,dbg_facility,"analize_leg: no seskey_match: %s\n",key);
#endif
		}
	}
	
	if (leg->ses) reanalyze_leg(leg);
}


void analize_pdu(thing_pdu* pdu) {
	/* TODO: 
    return a g_boolean to tell we've destroyed the pdu when the pdu is unnassigned
	destroy the unassigned pdu
	*/
	thing_leg* leg = NULL;
	guint8* leg_key;
	guint8* orig_leg_key = NULL;
	//thing_ses* ses = NULL;
	//void** cookie = NULL;
	AVPL* candidate_leg_key_match = NULL;
	AVPL* candidate_start = NULL;
	AVPL* candidate_stop = NULL;
	AVPL* is_start = NULL;
	AVPL* is_stop = NULL;
	AVPL* legkey_match = NULL;
#ifdef _THING_DEBUGGING
	guint8* avpl_str = NULL;
	
	dbg_print (dbg_leg,1,dbg_facility,"analize_pdu: %s\n",pdu->kind);
#endif

	apply_transforms(pdu->avpl, KEYWORD_PDU);
	
	candidate_leg_key_match = (AVPL*) g_hash_table_lookup(legkeys,pdu->kind);
	
	if (! candidate_leg_key_match) return;
#ifdef _THING_DEBUGGING
	avpl_str = avpl_to_str(candidate_leg_key_match);
	dbg_print (dbg_leg,1,dbg_facility,"analize_pdu: got candidate key: %s\n",avpl_str);
	g_free(avpl_str);
#endif
	
	legkey_match = new_avpl_exact_match("",pdu->avpl,candidate_leg_key_match, TRUE);
		
	if (legkey_match) {
		leg_key = avpl_to_str(legkey_match);

		candidate_start = (AVPL*) g_hash_table_lookup(starts,pdu->kind);

		if (candidate_start) {
#ifdef _THING_DEBUGGING
			avpl_str = avpl_to_str(candidate_start);
			dbg_print (dbg_leg,1,dbg_facility,"analize_pdu: got candidate start: %s\n",avpl_str);
			g_free(avpl_str);
#endif
			is_start = new_avpl_exact_match("",pdu->avpl, candidate_start, FALSE);
		}
		
		if (is_start) {
#ifdef _THING_DEBUGGING
			avpl_str = avpl_to_str(is_start);
			dbg_print (dbg_leg,1,dbg_facility,"analize_pdu: got start match: %s\n",avpl_str);
			g_free(avpl_str);
#endif
			delete_avpl(is_start,FALSE);	
		}
		
		g_hash_table_lookup_extended(sessions,leg_key,(gpointer*)&orig_leg_key,(gpointer*)&leg);
			
		if ( leg ) {
			g_free(leg_key);
			leg_key = orig_leg_key;
			
#ifdef _THING_DEBUGGING
			dbg_print (dbg_leg,1,dbg_facility,"analize_pdu: got leg: %s\n",leg_key);
			dbg_print (dbg_leak,2,dbg_facility,"analize_pdu: free: %X\n",leg_key);
#endif
			
			if (is_start) {
				if ( leg->released ) {
#ifdef _THING_DEBUGGING
					dbg_print (dbg_leg,1,dbg_facility,"analize_pdu: new leg on released key before key expiration\n");
#endif
					g_hash_table_remove(sessions,leg_key);
					leg = new_leg(pdu,leg_key);
					g_hash_table_insert(sessions,leg_key,leg);
				}
#ifdef _THING_DEBUGGING
				dbg_print (dbg_leg,1,dbg_facility,"analize_pdu: duplicate start on leg\n");
#endif
			}
			
			pdu->leg = leg;
			if (leg->last_pdu) leg->last_pdu->next = pdu;
			leg->last_pdu = pdu;
			pdu->next = NULL;
			pdu->rel_time -= leg->start_time;
			if (leg->released) pdu->after_release = TRUE;
			
		} else {
#ifdef _THING_DEBUGGING
			dbg_print (dbg_leg,1,dbg_facility,"analize_pdu: no leg\n");
#endif
			if (is_start) {
				leg = new_leg(pdu,leg_key);
				//g_message("analyze_pdu: got ney leg: %s",leg->id);
				g_hash_table_insert(sessions,leg_key,leg);
			} else {
#ifdef _THING_DEBUGGING
				dbg_print (dbg_leg,1,dbg_facility,"analize_pdu: an unassigned pdu\n");
#endif
				pdu->leg = NULL;
				pdu->next = NULL;
				/* TO DO:  destroy unassigned pdus immediately */
				return;
			}
		}
		
#ifdef _THING_DEBUGGING
		dbg_print (dbg_leg,4,dbg_facility,"analize_pdu: merge with key\n");
#endif
		merge_avpl(leg->avpl,legkey_match,TRUE);
		delete_avpl(legkey_match,TRUE);

#ifdef _THING_DEBUGGING
		dbg_print (dbg_leg,4,dbg_facility,"analize_pdu: apply extras\n");
#endif
		apply_extras(pdu->avpl,leg->avpl,pdu->kind);
		
#ifdef _THING_DEBUGGING
		dbg_print (dbg_leg,4,dbg_facility,"analize_pdu: apply transforms\n");
#endif
		
		/* do transforms for leg realy make sense ? */
		apply_transforms(leg->avpl,pdu->kind);

#ifdef _THING_DEBUGGING
		avpl_str = avpl_to_str(leg->avpl);
		dbg_print (dbg_leg,1,dbg_facility,"analize_pdu: Leg Attributes: %s\n",avpl_str);
		g_free(avpl_str);
#endif
		
		leg->last_time = pdu->rel_time;
		
		if ( ! leg->released) {
			candidate_stop = (AVPL*) g_hash_table_lookup(stops,pdu->kind);
			if (candidate_stop) {
#ifdef _THING_DEBUGGING
				dbg_print (dbg_leg,4,dbg_facility,"analize_pdu: got candidate stop\n");
#endif
				is_stop = new_avpl_exact_match("",pdu->avpl, candidate_stop,FALSE);
			}
			
			if(is_stop) {
#ifdef _THING_DEBUGGING
				avpl_str = avpl_to_str(is_stop);
				dbg_print (dbg_leg,1,dbg_facility,"analize_pdu: is_stop: %s\n",avpl_str);
				g_free(avpl_str);
#endif
				delete_avpl(is_stop,FALSE);
				
				if (! leg->released) {
					leg->released = TRUE;
					leg->release_time = pdu->rel_time;
					if (leg->ses) leg->ses->num_of_released_legs++;
				}
				
				pdu->is_stop = TRUE;
			} else {
#ifdef _THING_DEBUGGING
				dbg_print (dbg_leg,4,dbg_facility,"analize_pdu: is not a stop\n");
#endif
			}
		}
		
		if (leg->ses) {
			reanalyze_leg(leg);
		} else {
			analize_leg(leg);
		}
		
	} else {
#ifdef _THING_DEBUGGING
		dbg_print (dbg_leg,4,dbg_facility,"analize_pdu: no leg_key\n");
#endif
		
		/* pdu->expire = insert_expire( pdu->t + lifetime ,pdu,NULL); */
		pdu->leg = NULL;
		/* TO DO:  destroy unassigned pdus immediately */

		return;
	}
	
}

thing_pdu* do_thing(tvbuff_t *tvb, proto_tree* tree, packet_info* pinfo) {
	working_packet wp;
	thing_pdu *pdus = NULL;
	guint32* fnp = NULL;
	thing_pdu *pdu = NULL;
	thing_pdu *next = NULL;
	thing_pdu *last = NULL;
	
#ifdef _THING_DEBUGGING
	dbg_print (dbg,2,dbg_facility,"do_thing: got frame number: %i\n",pinfo->fd->num);
#endif

	tvb = NULL; /* no warning: should I get rid of it? */
	
	pdus = (thing_pdu*) g_hash_table_lookup(visited_packets,&pinfo->fd->num);
	
	if (pdus) return pdus;

	/* if it's lower than the highest_analyzed_frame and isn't in the hash
		that means that there's no information on that frame anymore */
	
	/* if (pinfo->fd->num <= highest_analyzed_frame) return NULL; */

	if ( tree->tree_data && tree->tree_data->interesting_hfids ) {
#ifdef _THING_DEBUGGING
		dbg_print (dbg,2,dbg_facility,"do_thing: frame with tree not seen before: %i\n",pinfo->fd->num);
#endif
		
		wp.transport.finfo = NULL;
		wp.transport.next = &wp.transport;
		wp.transport.prev = &wp.transport;
		wp.pdu.finfo = NULL;
		wp.pdu.next = &wp.pdu;
		wp.pdu.prev = &wp.pdu;
		wp.attr.finfo = NULL;
		wp.attr.next = &wp.attr;
		wp.attr.prev = &wp.attr;
		wp.framenum = pinfo->fd->num;
		wp.t = pinfo->fd->rel_secs;
		wp.ut = pinfo->fd->rel_usecs;

#ifdef _THING_DEBUGGING
		dbg_print (dbg,6,dbg_facility,"do_thing: about to get finfos\n");
#endif
		
		/* TO DO:
		 *   leave analysis sooner whenever there is nothing worth.
		 *   I should try to look for protos before anything else,
		 *   the current approach of importing everything and then
		 *   deciding whether we need or not to create pdus is plain
		 *   wrong. Anyway the whole interface to ethereal sucks.
		 *   It has to be rewritten!
		 */
		
		extract_interesting_finfos(tree,&wp);
		
#ifdef _THING_DEBUGGING
		dbg_print (dbg,5,dbg_facility,"do_thing: got finfos\n");
		if (*dbg > 5) print_wp(&wp);
#endif		
		
		pdus = make_pdus(&wp);
		
		cleanup_working_packet(&wp);
		
		for (pdu = pdus; pdu; pdu = next) {
			next = pdu->next_in_frame;
			
			analize_pdu(pdu);
			
#if 0
			
			/* should this be done here or in analyze_pdu ? */
			if (! pdu->leg ) {
				/* there is no need to keep unassigned pdus */

				if (last) {
					last->next_in_frame = next;
				}
				
				if ( pdu == pdus ) {
					pdus = next;
				}
				
				pdu->next_in_frame = NULL;
				
				delete_thing(pdu);
				
			} else if (! keep_pdu_attributes ) {
				/* there is no need to keep pdu's avps */
				delete_avpl(pdu->avpl,TRUE);
				pdu->avpl = NULL;
			}
			
			last = pdu;
#endif

		}
		
#ifdef _THING_DEBUGGING
		dbg_print (dbg,5,dbg_facility,"do_thing: make_pdus done\n");
#endif
		/* TODO: I got to handle the pdu->frame relation better! */
		if (pdus) {
			fnp = g_malloc(sizeof(guint32));
			*fnp = pinfo->fd->num;
			g_hash_table_insert(visited_packets,fnp,pdus);
			last_visited_frame = pinfo->fd->num;
		}
		
		/* highest_analyzed_frame = pinfo->fd->num; */
		
	}
	// //g_message("make_pdus done");
	return pdus;
}

/* debug code */
#ifdef _THING_DEBUGGING
void print_hfid_hash(gpointer k, gpointer v, gpointer p) {
	dbg_print (dbg,5,dbg_facility,"print_hfid_hash: %s(%i)=%s",(guint8*)v,*(int*)k,(guint8*)p);
}
#endif

/* for cleaning hashes */
gboolean free_both(gpointer k, gpointer v, gpointer p) {
	g_free(k);
	if (p) g_free(v);
	return TRUE;
}

/* for cleaning hashes of ptr arrays */
gboolean free_GPtrArray_value(gpointer k, gpointer v, gpointer p)
{
	g_free(k);
	
	if (p) {
		GPtrArray   *ptrs = v;
		g_ptr_array_free(ptrs, TRUE);
	}

	return TRUE;
}


gboolean add_hfid(guint8* what, guint8* how, GHashTable* where) {
	header_field_info*  hfi = NULL;
	header_field_info*  first_hfi = NULL;
	gboolean exists = FALSE;
	guint8* as;
	guint8* h;
	int* ip;
	
	hfi = proto_registrar_get_byname(what);
	
	while(hfi) {
		first_hfi = hfi;
		hfi = hfi->same_name_prev;
	}
	
	hfi = first_hfi;
	
	while (hfi) {
		exists = TRUE;
		ip = g_malloc(sizeof(int));
		*ip = hfi->id;
		
		
		if (( as =g_hash_table_lookup(where,ip) )) {
			g_free(ip);
			if (! g_str_equal(as,how)) {
				g_warning("Error: add_hfid: attempt to add %s(%i) as %s"
						  " failed: hfid already added as '%s'",what,hfi->id,how,as);
				return FALSE;
			}
		} else {
			g_ptr_array_add(thing_hfids,ip);
			h = g_strdup(how);
			g_hash_table_insert(where,ip,h);
			
			
#ifdef _THING_DEBUGGING
			dbg_print (dbg,5,dbg_facility,"add_hfid: added hfid %s(%i) as %s",what,*ip,how);
#endif
		}
		
		hfi = hfi->same_name_next;
		
	}
	
	if (! exists) {
		g_warning("Error: add_hfid(%s): cannot find",what);
	}
	
	return exists;
}

/* a configuration error */
void thing_config_error(LoAL* loal ,guint8* line , guint8* fmt, ...) {
	va_list list;
	guint8* desc;
	
	
	va_start( list, fmt );
	desc = g_strdup_vprintf(fmt, list);
	va_end( list );
	
#ifdef _THING_DEBUGGING
	dbg_print (dbg,0,dbg_facility,"thing_config_error: %s",desc);
#endif

	g_warning(desc);
	g_free(desc);
	
	disable_thing();
	/* delete_loal(loal,TRUE,TRUE); */
	if ( line ) g_free(line);
}

gboolean thing_make_config(LoAL* loal) {
	AVPL* avpl;
	AVP* action_avp = NULL;
	config_action* action;
	guint8* line = NULL;
	
	if (! loal ) {
		thing_config_error(loal,line,"thing: Error: no config!");
		return FALSE;
	}
	
	
	while(( avpl = extract_first_avpl(loal) )) {
		line = avpl_to_str(avpl);
		
#ifdef _THING_DEBUGGING
		dbg_print (dbg_cfg,2,dbg_facility,"thing_make_config: current line: %s",line);
#endif
		
		action_avp = extract_avp_by_name(avpl,KEYWORD_ACTION);
		
		if (! action_avp ) {
			thing_config_error(loal,line,"thing: Config error: no action in '%s'",line);
			return FALSE;
		}
		
		action = g_hash_table_lookup(actions,action_avp->v);
		
		if (action) {
			if ( ! action(avpl) ) {
				thing_config_error(loal,line,"thing: Failed: action '%s' on '%s'",action_avp->v,line);
				return FALSE;
			}
		} else {
			thing_config_error(loal,line,"thing: Error: action '%s' unknown in '%s'",action_avp->v,line);
			return FALSE;
		}
		
		delete_avp(action_avp);

		
		g_free(line);
	}
	
#ifdef _THING_DEBUGGING
	dbg_print (dbg_cfg,2,dbg_facility,"thing_make_config: done");
#endif
	
	return TRUE;
}


gboolean free_avpl(gpointer k, gpointer v, gpointer p) {
	AVPL* avpl = (AVPL*) v;

	p = k;
	
	g_free(k);
	delete_avpl(avpl,TRUE);
	return TRUE;
}

gboolean free_loal(gpointer k, gpointer v, gpointer p) {
	LoAL* loal = (LoAL*) v;
	
	p = k;
	
	g_free(k);
	delete_loal(loal,TRUE,TRUE);
	return TRUE;
}

gboolean free_none(gpointer k, gpointer v, gpointer p) { k = v = p; return TRUE; }

gboolean free_transform(gpointer k, gpointer v, gpointer p) {
	GPtrArray* ptrs = (GPtrArray*) v;
	guint i;
	
	v = p; /* no warning */
	
	g_free(k);
	
	for (i=0;i < ptrs->len; i++) {
		delete_avpl_transform((AVPL_Transf*) g_ptr_array_index(ptrs,i));
	}
	
	g_ptr_array_free(ptrs,FALSE);
	return TRUE;

}

void init_avpl_hash(GHashTable** hash) {
	if (*hash) {
		g_hash_table_foreach_remove(*hash,free_avpl,NULL);
	} else {
		*hash = g_hash_table_new(g_str_hash,g_str_equal);
	}
}

#define VALUE_TOO ((void*)1)

void init_hfid_hash(GHashTable** hash) {
	if (*hash) {
		g_hash_table_foreach_remove(*hash,free_both,VALUE_TOO); 
	} else {
		*hash = g_hash_table_new(g_int_hash,g_int_equal);
	}
}

#ifdef _THING_DEBUGGING
void print_transforms_hash(gpointer k, gpointer v, gpointer p) {
	GPtrArray* ptrs = (GPtrArray*) v;
	AVPL_Transf* t;
	guint8* match = NULL;
	guint8* replace = NULL;
	guint8* match_op = NULL;
	guint8* replace_op = NULL;
	guint i;
	
	//g_message("print_transforms_hash for: %s, %X[%i]",k,ptrs,ptrs->len);
	for (i=0; i<ptrs->len; i++) {
		t = g_ptr_array_index(ptrs,i);
		
		
		if (t->match) match = avpl_to_str(t->match);
		if (t->replace) replace = avpl_to_str(t->replace);
		
		switch (t->match_mode) {
			case EXACT:
				match_op = "exact";
				break;
			case LOOSE:
				match_op = "loose";
				break;
			case EVERY: 
				match_op = "every";
				break;
			case NO_MATCH:
				match_op = "no_match";
				break;
		}
		
		switch (t->replace_mode) {
			case NO_REPLACE:
				replace_op = "no_replace";
				break;
			case INSERT:
				replace_op = "insert";
				break;
			case REPLACE:
				replace_op = "replace";
				break;
		}
		
		
		dbg_print (dbg_cfg,1,dbg_facility,
				   "%s: %s[%i]  %s(%s) ==> %s(%s)",
				   (guint8*)p,(guint8*)k,i+1,match_op,match,replace_op,replace);
		
		g_free(match);
		g_free(replace);
	}
}

void print_avpl_hash(gpointer k, gpointer v, gpointer p) {
	AVPL* avpl = (AVPL*) v;
	guint8* s = avpl_to_str(avpl);
	
	dbg_print (dbg_cfg,1,dbg_facility,"%s: '%s' >> %s",(guint8*)p,(guint8*)k,s);
	
	g_free(s);
}

void print_config() {
	g_hash_table_foreach(legkeys,print_avpl_hash,"legkeys");
	g_hash_table_foreach(starts,print_avpl_hash,"starts");
	g_hash_table_foreach(stops,print_avpl_hash,"stops");
	g_hash_table_foreach(extras,print_avpl_hash,"extras");
	//g_hash_table_foreach(seskeys,print_loal_hash,NULL);
	g_hash_table_foreach(transforms,print_transforms_hash,"transforms");
}
#endif

gboolean config_include(AVPL*avpl) {
	void* cookie = NULL;
	AVP* file_avp = get_avp_by_name(avpl,KEYWORD_FILENAME,&cookie);
	guint8* filename;
	
	
	if(file_avp) {
		filename = file_avp->v;
	} else {
		thing_config_error(NULL,NULL,"thing: Include file error: no filename");
		return FALSE;
	}
	
	/* FIXME: stop recursion */
	if ( ! thing_make_config(loal_from_file(filename)) ) {
		thing_config_error(NULL,NULL,"thing: Error Loading '%s'",filename);
		return FALSE;
	} 
	
	delete_avp(file_avp);
	
	return TRUE;
}



gboolean config_debug(AVPL*avpl) {
#ifdef _THING_DEBUGGING
	void* cookie = NULL;
	AVP* avp;
#ifdef _AVP_DEBUGGING
	int debug_avp = 0;
	int dbg_avp = 0;
	int dbg_avp_op = 0;
	int dbg_avpl = 0;
	int dbg_avpl_op = 0;
#endif

	if(( avp = get_avp_by_name(avpl,KEYWORD_FILENAME,&cookie) )) {
		dbg_facility = fopen(avp->v,"w");
		delete_avp(avp);
		avp = NULL;
	}
	
	cookie = NULL;
	
#define CONFIG_DEBUG(keyword,var_name) if(( avp = get_avp_by_name(avpl,keyword,&cookie) )) { var_name = strtol(avp->v,NULL,10); delete_avp(avp); avp = NULL;} cookie = NULL
	
	CONFIG_DEBUG(KEYWORD_DBG_GENERAL,dbg_lvl);
	CONFIG_DEBUG(KEYWORD_DBG_CFG,dbg_cfg_lvl);
	CONFIG_DEBUG(KEYWORD_DBG_LEAK,dbg_leak_lvl);
	CONFIG_DEBUG(KEYWORD_DBG_PDU,dbg_pdu_lvl);
	CONFIG_DEBUG(KEYWORD_DBG_LEG,dbg_leg_lvl);
	CONFIG_DEBUG(KEYWORD_DBG_SES,dbg_ses_lvl);
	
#ifdef _AVP_DEBUGGING
	CONFIG_DEBUG(KEYWORD_DBG_AVPLIB,debug_avp);
	CONFIG_DEBUG(KEYWORD_DBG_AVP,dbg_avp);
	CONFIG_DEBUG(KEYWORD_DBG_AVP_OP,dbg_avp_op);
	CONFIG_DEBUG(KEYWORD_DBG_AVPL,dbg_avpl);
	CONFIG_DEBUG(KEYWORD_DBG_AVPL_OP,dbg_avpl_op);

	setup_avp_debug(dbg_facility, debug_avp, dbg_avp, dbg_avp_op, dbg_avpl, dbg_avpl_op)
#endif
		
#endif
		return TRUE;
}


/* a list transform */
gboolean config_transform(AVPL* avpl) {
	AVP* on_avp = extract_avp_by_name(avpl,KEYWORD_ON);
	AVP* method_avp = extract_avp_by_name(avpl,KEYWORD_METHOD);
	avpl_match_mode match_mode = NO_MATCH;
	avpl_replace_mode replace_mode = NO_REPLACE;
	AVPL_Transf* t;
	GPtrArray* ptrs;
	guint8* key;
	
	if (! on_avp ) {
		thing_config_error(NULL, NULL, "thing: Config error: no on field in transform");
		return FALSE;
	}
	
	if (! method_avp ) {
		delete_avp(on_avp);
		thing_config_error(NULL, NULL, "thing: Config error: no method field in transform");
		return FALSE;
	}
	
	if (g_str_equal(method_avp->v,KEYWORD_EXACT)) {
		match_mode = EXACT;
	} else if (g_str_equal(method_avp->v,KEYWORD_LOOSE)) {
		match_mode = LOOSE;
	} else if (g_str_equal(method_avp->v,KEYWORD_EVERY)) {
		match_mode = EVERY;
	} else if (g_str_equal(method_avp->v,KEYWORD_INSERT)) {
		replace_mode = INSERT;
	} else if (g_str_equal(method_avp->v,KEYWORD_REPLACE)) {
		replace_mode = REPLACE;
	}
	
	if (match_mode == NO_MATCH && replace_mode == NO_REPLACE) {
		thing_config_error(NULL, NULL, "thing: Config error: wrong method field in transform: '%s'",method_avp->v);
		delete_avp(on_avp);
		delete_avp(method_avp);
		return FALSE;
	}
	
	if (( ptrs =  g_hash_table_lookup(transforms,on_avp->v) )) {
		t = g_ptr_array_index(ptrs,ptrs->len - 1);
		
		if ( t->match && t->replace ) {
			t = new_avpl_transform();
			g_ptr_array_add(ptrs,t);
		}			
	} else {
		ptrs = g_ptr_array_new();
		
		t = new_avpl_transform();
		g_ptr_array_add(ptrs,t);
		
		key = g_strdup(on_avp->v);
		g_hash_table_insert(transforms,key,ptrs);
	}
	
	/* I assume match and replace come one after another 
		weird things can happen if it is not exactly that way */
	
	if ( match_mode  != NO_MATCH ) {
		set_avpl_transform_match(t, match_mode, avpl);
	} 
	
	if ( replace_mode != NO_REPLACE ) {
		set_avpl_transform_replace(t, replace_mode, avpl);
	}
	
	delete_avp(on_avp);
	delete_avp(method_avp);
	return TRUE;
}


gboolean config_pdu(AVPL* avpl) {
	AVP* transport_avp;
	AVP* pdu_avp;
	AVP* attr_avp;
	guint8* line = NULL;
	
	line = avpl_to_str(avpl);
	
	
#ifdef _THING_DEBUGGING
	dbg_print (dbg_pdu,3,dbg_facility,"config_pdu: pdukey: %s",line);
#endif
	
	transport_avp = extract_avp_by_name(avpl,KEYWORD_TRANSPORT);
	
	if (! transport_avp ) {
		g_warning("thing: PDU Config error: no transport in '%s'",line);
		g_free(line);
		delete_avpl(avpl,TRUE);
		return FALSE;
	}
	
	if ( ! add_hfid(transport_avp->v,transport_avp->v,hfids_transport) ) {
		g_warning("thing: PDU Config error: failed to set transport '%s' in '%s'",transport_avp->v,line);
		g_free(line);
		delete_avpl(avpl,TRUE);
		return FALSE;
	}
	
	delete_avp(transport_avp);
	
	pdu_avp = extract_avp_by_name(avpl,KEYWORD_PROTO);
	
	if (! pdu_avp ) {
		g_warning("thing: PDU Config error: no proto in '%s'",line);
		g_free(line);
		delete_avpl(avpl,TRUE);
		return FALSE;
	}
	
	if ( ! add_hfid(pdu_avp->v,pdu_avp->v,hfids_pdu) ) {
		g_warning("thing: PDU Config error: failed to set proto '%s' in '%s'",pdu_avp->v,line);
		g_free(line);
		delete_avpl(avpl,TRUE);
		return FALSE;
	}
	
	delete_avp(pdu_avp);
	
	while (( attr_avp = extract_first_avp(avpl) )) {
		if ( ! add_hfid(attr_avp->v,attr_avp->n,hfids_attr) ) {
			g_warning("thing: PDU Config error: failed to set attr '%s' in '%s'",attr_avp->n,line);
			g_free(line);
			delete_avp(attr_avp);
			delete_avpl(avpl,TRUE);
			return FALSE;
		}
	}
	
	delete_avpl(avpl,TRUE);
	return TRUE;
}

AVPL* all_keywords = NULL;
/* not used keywords should not disappear from avp scs */

void init_actions() {
	
	if (! all_keywords ) {
		all_keywords = new_avpl("all_keywords");
		insert_avp(all_keywords,new_avp(KEYWORD_ACTION,"",'='));
		insert_avp(all_keywords,new_avp(KEYWORD_SETTINGS,"",'='));
		insert_avp(all_keywords,new_avp(KEYWORD_INCLUDE,"",'='));
		insert_avp(all_keywords,new_avp(KEYWORD_PDU,"",'='));
		insert_avp(all_keywords,new_avp(KEYWORD_TRANSFORM,"",'='));
		insert_avp(all_keywords,new_avp(KEYWORD_LEGKEY,"",'='));
		insert_avp(all_keywords,new_avp(KEYWORD_LEGSTART,"",'='));
		insert_avp(all_keywords,new_avp(KEYWORD_LEGSTOP,"",'='));
		insert_avp(all_keywords,new_avp(KEYWORD_LEGEXTRA,"",'='));
		insert_avp(all_keywords,new_avp(KEYWORD_SESKEY,"",'='));
		insert_avp(all_keywords,new_avp(KEYWORD_SESEXTRA,"",'='));
		insert_avp(all_keywords,new_avp(KEYWORD_ON,"",'='));
		insert_avp(all_keywords,new_avp(KEYWORD_FILENAME,"",'='));
		insert_avp(all_keywords,new_avp(KEYWORD_NAME,"",'='));
		insert_avp(all_keywords,new_avp(KEYWORD_PROTO,"",'='));
		insert_avp(all_keywords,new_avp(KEYWORD_METHOD,"",'='));
		insert_avp(all_keywords,new_avp(KEYWORD_TRANSFORM,"",'='));
		insert_avp(all_keywords,new_avp(KEYWORD_METHOD,"",'='));
		insert_avp(all_keywords,new_avp(KEYWORD_EXACT,"",'='));
		insert_avp(all_keywords,new_avp(KEYWORD_LOOSE,"",'='));
		insert_avp(all_keywords,new_avp(KEYWORD_EVERY,"",'='));
		insert_avp(all_keywords,new_avp(KEYWORD_REPLACE,"",'='));
		insert_avp(all_keywords,new_avp(KEYWORD_INSERT,"",'='));
		insert_avp(all_keywords,new_avp(KEYWORD_MAP,"",'='));
		insert_avp(all_keywords,new_avp(KEYWORD_SESEXPIRE,"",'='));
		insert_avp(all_keywords,new_avp(KEYWORD_DISCARDPDU,"",'='));
		insert_avp(all_keywords,new_avp(KEYWORD_DROPPDU,"",'='));
		insert_avp(all_keywords,new_avp(KEYWORD_LIBPATH,"",'='));
		insert_avp(all_keywords,new_avp(KEYWORD_DROPLEG,"",'='));
		insert_avp(all_keywords,new_avp(KEYWORD_FROM,"",'='));
		insert_avp(all_keywords,new_avp(KEYWORD_TO,"",'='));
		
#ifdef _THING_DEBUGGING
		insert_avp(all_keywords,new_avp(KEYWORD_DBG,"",'='));
		insert_avp(all_keywords,new_avp(KEYWORD_DBG_GENERAL,"",'='));
		insert_avp(all_keywords,new_avp(KEYWORD_DBG_LEAK,"",'='));
		insert_avp(all_keywords,new_avp(KEYWORD_DBG_CFG,"",'='));
		insert_avp(all_keywords,new_avp(KEYWORD_DBG_PDU,"",'='));
		insert_avp(all_keywords,new_avp(KEYWORD_DBG_LEG,"",'='));
		insert_avp(all_keywords,new_avp(KEYWORD_DBG_SES,"",'='));
#ifdef _AVP_DEBUGGING
		insert_avp(all_keywords,new_avp(KEYWORD_DBG_AVPLIB,"",'='));		
		insert_avp(all_keywords,new_avp(KEYWORD_DBG_AVP,"",'='));
		insert_avp(all_keywords,new_avp(KEYWORD_DBG_AVP_OP,"",'='));		
		insert_avp(all_keywords,new_avp(KEYWORD_DBG_AVPL,"",'='));
		insert_avp(all_keywords,new_avp(KEYWORD_DBG_AVPL_OP,"",'='));
#endif
#endif
		
	}
	
	if (actions) {
		g_hash_table_destroy(actions);
	}
	
	actions = g_hash_table_new(g_str_hash,g_str_equal);
	
}

void new_action(guint8* name, config_action* action) {
	g_hash_table_insert(actions,name,action);
}

/* FIXME: even if handy these two are REALY ugly */

#define ACTION_AVPL_TO_HASH(the_name, the_hash)\
gboolean the_name(AVPL* avpl) {														\
	AVP* on_avp; guint8* line = NULL; guint8* key;									\
		line = avpl_to_str(avpl); on_avp = extract_avp_by_name(avpl,KEYWORD_ON);				\
			if (! on_avp ) { g_warning("thing: PDU Config error: no 'on' in '%s'",line);	\
				g_free(line); delete_avpl(avpl,TRUE); return FALSE; }						\
			key = g_strdup(on_avp->v); delete_avp(on_avp);									\
				g_hash_table_insert(the_hash,key,avpl); g_free(line); return TRUE; }


ACTION_AVPL_TO_HASH(config_legkey,legkeys);
ACTION_AVPL_TO_HASH(config_start,starts);
ACTION_AVPL_TO_HASH(config_stop,stops);
ACTION_AVPL_TO_HASH(config_legextra,extras);
ACTION_AVPL_TO_HASH(config_sesextra,extras);


gboolean config_seskey(AVPL* avpl) { 
	LoAL* straight;
	LoAL* reverse;
	AVP* on_avp;
	AVP* name_avp;
	guint8* key;
	
	on_avp = extract_avp_by_name(avpl,KEYWORD_ON);
	
	if (! on_avp ) {
		g_warning("thing: SesKey Config error: no " KEYWORD_ON);
		return FALSE;
	}
	
	name_avp = extract_avp_by_name(avpl,KEYWORD_NAME);
	
	if (! name_avp ) {
		g_warning("thing: SesKey Config error: no " KEYWORD_NAME);
		delete_avp(on_avp);
		return FALSE;
	}
	
	rename_avpl(avpl,name_avp->v);
		
	straight = (LoAL*) g_hash_table_lookup(seskeys,name_avp->v);
	
	if (! straight) {
		key = g_strdup(name_avp->v);
		straight = new_loal(on_avp->v);
		g_hash_table_insert(seskeys,key,straight);
	} 
	
	loal_append(straight,avpl);
		
	reverse  = (LoAL*) g_hash_table_lookup(syekses,on_avp->v);
	
	if ( ! reverse ) {
		key = g_strdup(on_avp->v);
		reverse = new_loal(name_avp->v);
		g_hash_table_insert(syekses,key,reverse);
	}
	
	loal_append(reverse,avpl);

	delete_avp(on_avp);
	delete_avp(name_avp);
	
	return TRUE;
}

extern void initialize_thing(guint8* configuration_filename) {

#ifdef _THING_DEBUGGING
	dbg_print (dbg,5,dbg_facility,"initialize_thing: entering");
#endif
	
	avp_init( (max_things ? max_things : 100) * workspace_scale);
	scs_init(&strings);

	init_actions();
	new_action(KEYWORD_DBG,config_debug);
	new_action(KEYWORD_PDU,config_pdu);
	new_action(KEYWORD_LEGKEY,config_legkey);
	new_action(KEYWORD_SESKEY,config_seskey);
	new_action(KEYWORD_LEGSTART,config_start);
	new_action(KEYWORD_LEGSTOP,config_stop);
	new_action(KEYWORD_LEGEXTRA,config_legextra);
	new_action(KEYWORD_SESEXTRA,config_sesextra);
	new_action(KEYWORD_INCLUDE,config_include);
	new_action(KEYWORD_TRANSFORM,config_transform);

	
	if (things) {
		g_hash_table_foreach_remove(things,free_both,NULL);
	} else {
		things = g_hash_table_new(g_str_hash,g_str_equal);
	}
	
	
	/* config */
	init_hfid_hash(&hfids_transport);
	init_hfid_hash(&hfids_pdu);
	init_hfid_hash(&hfids_attr);
		
	init_avpl_hash(&legkeys);
	init_avpl_hash(&starts);
	init_avpl_hash(&stops);
	init_avpl_hash(&extras);
	
	if (transforms) {
		g_hash_table_foreach_remove(transforms,free_transform,NULL);
	} else {
		transforms = g_hash_table_new(g_str_hash,g_str_equal);
	}
	
	if (seskeys) {
		g_hash_table_foreach_remove(seskeys,free_loal,NULL);
	} else {
		seskeys = g_hash_table_new(g_str_hash,g_str_equal);
	}

	if (syekses) {
		g_hash_table_foreach_remove(syekses,free_loal,NULL);
	} else {
		syekses = g_hash_table_new(g_str_hash,g_str_equal);
	}
	
	if (thing_hfids) {
		g_ptr_array_free(thing_hfids,TRUE);
	} 
	
	thing_hfids = g_ptr_array_new();
	
	/* runtime */
	if (visited_packets) {
		g_hash_table_foreach_remove(visited_packets,free_both,NULL);
	} else {
		visited_packets = g_hash_table_new(g_int_hash,g_int_equal);
	}
	
	if (sessions) {
		g_hash_table_foreach_remove(sessions,free_both,NULL);
	} else {
		sessions = g_hash_table_new(g_str_hash,g_str_equal);
	}
	
	if (sessids) {
		g_hash_table_foreach_remove(sessids,free_both,NULL);
	} else {
		sessids = g_hash_table_new(g_str_hash,g_str_equal);
	}
	
	if (circle_of_life_initialized) {
		thing_flush(); /* this will clear all existing runtime data */
	} else {
		memset(&circle_of_life, '\0', sizeof(thing));
		circle_of_life.next_in_ring = &circle_of_life;
		circle_of_life.prev_in_ring = &circle_of_life;
	}

	
	thing_make_config(loal_from_file(configuration_filename));
	
	//g_hash_table_foreach(hfids_transport,print_hfid_hash,"transport");
	//g_hash_table_foreach(hfids_pdu,print_hfid_hash,"pdu");
	//g_hash_table_foreach(hfids_attr,print_hfid_hash,"attr");
	
#ifdef _THING_DEBUGGING
	if (*dbg_cfg > 0) print_config();
#endif
	
}

extern void thing_prime_edt(epan_dissect_t* edt) {	
	guint i;
	int hfid;
	
	if (! edt->tree)  return;

	for (i=0; i < thing_hfids->len; i++) {
		hfid = *((int*) g_ptr_array_index(thing_hfids,i));
		proto_tree_prime_hfid(edt->tree,hfid);
		
#ifdef _THING_DEBUGGING
		dbg_print (dbg,4,dbg_facility,"thing_prime_edt: hfid: %i",hfid);
#endif
		
	}
}

extern void thing_keep_pdu_attributes() {;}
extern void thing_discard_pdu_attributes(){;}


