/* thing.h
* The session tracing thing 
*
* Copyright 2004, Luis E. Garcia Ontanon <luis.ontanon@gmail.com>
*
* $Id: thing.h,v 0.4 2004/11/14 03:28:47 lego Exp $
*
* Ethereal - Network traffic analyzer
* By Gerald Combs <gerald@ethereal.com>
* Copyright 1998 Gerald Combs
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
* 
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/


#ifndef __THING_H_
#define __THING_H_

#include "config.h"
#include <epan/thing_util.h>
#include <epan/epan_dissect.h>
#include <epan/packet.h>
#include <glib.h>

/* #define _THING_DEBUGGING */

typedef gboolean config_action(AVPL* avpl);

/* Config AVP Names */
#define KEYWORD_ACTION "Action"
#define KEYWORD_SETTINGS "Settings"
#define KEYWORD_INCLUDE "Include"
#define KEYWORD_PDU "PDU"
#define KEYWORD_TRANSFORM "Transform"
#define KEYWORD_LEGKEY "LegKey"
#define KEYWORD_LEGSTART "LegStart"
#define KEYWORD_LEGSTOP "LegStop"
#define KEYWORD_LEGEXTRA "LegExtra"
#define KEYWORD_SESKEY "SesKey"
#define KEYWORD_SESEXTRA "SesExtra"
#define KEYWORD_ON "On"
#define KEYWORD_FROM "From"
#define KEYWORD_TO "To"
#define KEYWORD_FILENAME "Filename"
#define KEYWORD_NAME "Name"
#define KEYWORD_PROTO "Proto"
#define KEYWORD_METHOD "Method"
#define KEYWORD_TRANSPORT "Transport"
#define KEYWORD_METHOD "Method"
#define KEYWORD_EXACT "Exact"
#define KEYWORD_LOOSE "Loose"
#define KEYWORD_EVERY "Every"
#define KEYWORD_REPLACE "Replace"
#define KEYWORD_INSERT "Insert"
#define KEYWORD_MAP "Map" /* not used (yet) */
#define KEYWORD_SESEXPIRE "SessionExpiration" /* not used (anymore) */
#define KEYWORD_DISCARDPDU "DiscardPduData" /* not used (yet) */
#define KEYWORD_DROPPDU "DiscardUnassignedPdus" /* not used (yet) */
#define KEYWORD_DROPLEG "DiscardUnassignedLegs" /* not used (yet) */
#define KEYWORD_LIBPATH "ThingLibPath" /* not used (yet) */
#define KEYWORD_DBG "Debug"

#ifdef _THING_DEBUGGING
#define KEYWORD_DBG_GENERAL "Debug_General"
#define KEYWORD_DBG_LEAK "Debug_Leak"
#define KEYWORD_DBG_CFG "Debug_Cfg"
#define KEYWORD_DBG_PDU "Debug_PDU"
#define KEYWORD_DBG_LEG "Debug_Leg"
#define KEYWORD_DBG_SES "Debug_Ses"
#ifdef _AVP_DEBUGGING
#define KEYWORD_DBG_AVPLIB "Debug_AVP_Lib"
#define KEYWORD_DBG_AVP "Debug_AVP"
#define KEYWORD_DBG_AVP_OP "Debug_AVP_Op"
#define KEYWORD_DBG_AVPL "Debug_AVPL"
#define KEYWORD_DBG_AVPL_OP "Debug_AVPL_Op"
#endif
#endif


typedef struct _thing thing;
typedef struct _thing thing_pdu;
typedef struct _thing thing_leg;
typedef struct _thing thing_ses;


/* these are used to contain information regarding pdus, legs and sess */
struct _thing {
	/* all three of them */
	guint8* id; /* who I am? */
	AVPL* avpl; /* the attributes of the pdu/leg/ses */
	guint8* kind;  /* whick kind this item is? */
	gint start; /* start of the pdu in the tvb / framenum of the start of a leg/ses */
	gint end;  /* end of the pdu in the tvb / framenul of the stop of a leg/ses */
	guint32 framenum; /* the frame from where it comes from */
	thing* next; /* pdu: next in leg, leg: next in ses, ses: ??? */
	gboolean released; /* has this leg been released? */
	float expiration; /* when will it expire once released? */
	
	
	enum _thing_type {THING_PDU_TYPE,THING_LEG_TYPE,THING_SES_TYPE} type;
	
	/* union _payload { */
		/* struct _pdu { */
			guint8* proto;
			thing_leg* leg; /* the leg the pdu belongs to (if any) */
			gboolean first; /* is this the first pdu in this frame? */
			gboolean is_start; /* this is the start pdu for this leg */
			gboolean is_stop; /* this is the stop pdu for this leg */
			gboolean after_release; /* this pdu comes after the stop */
			float rel_time; /* time since leg start if in leg or start of capture if unassigned */
			thing_pdu* next_in_frame; /* points to the next pdu in this frame */
		/* } pdu; */
		
		/* struct _leg { */
			thing_ses* ses; /* the ses-session of a leg */
			thing_pdu* pdus; /* pdus that belong to a leg (NULL in ses) */
			int num_of_pdus; /* how many legs a ses has? */
			int num_of_after_release_pdus;  /* how many pdus have arrived since it's been released */
			float start_time; /* time of start */
			float release_time; /* when this leg was released */
			float last_time; /* the rel time at which the last pdu/leg has been added */
			guint8* leg_key; /* used by leg */
			thing_pdu* last_pdu; /* last pdu in pdu's list */
		/* } leg; */
		
		/* struct _ses { */
			thing_leg* legs; /* legs that belong to a ses (NULL in leg) */
			int num_of_legs; /* how many legs a ses has? */
			int num_of_released_legs;  /* how many of them have already been released */
			guint last_n; /* the number of attributes the avpl had the last time we checked */
			GPtrArray* ses_keys; /* the keys under which this ses is stored in the sess hash */
			thing_leg* last_leg; /* last leg in leg's list */
		/* } ses; */
	/* } o; */
	
	thing* next_in_ring;
	thing* prev_in_ring;
	
};

extern gboolean thing_enabled();
extern void enable_thing();
extern void disable_thing();

extern void initialize_thing(guint8* configuration_filename);
extern void thing_prime_edt(epan_dissect_t* edt);

thing_pdu* do_thing(tvbuff_t *tvb,proto_tree* tree, packet_info* pinfo);

extern void thing_set_after_release_timer(guint32 seconds);
extern void thing_set_lifetime(guint32 seconds);
extern void thing_enable_expiration();
extern void thing_disable_expiration();
extern void thing_keep_pdu_attributes();
extern void thing_discard_pdu_attributes();

#endif
